/* tcc.c - Small-C compilation sequencer
 * -------------------------------------
 * Generate *exec (*.bat) file to compile Small-C programs
 */

/* 26-Jan-1989  v0.70 - A.J.Travis
 * 27-Jun-1991  v0.71 - J.G.Harston
 *   gets system paths from include.h
 *   uses '/' instead of '_' in filenames
 *   'stacks' execfiles for chained batchfiles
 *   if not chaining multiple batchfiles, deletes "compile" at end
 *   main() ends with exit(0);, paths increased from 80 to 128 chars
 */

#include <stdio.h>

#ifdef MSDOS
#define TCPP	"tcpp"
#define TCCOM	"tccom"
#define TCCOMC	"tccom -C"
#define AS65	"as65 -o"
#define RM	"rm -f"
#define LD	"ld -o"
#define SWLD	"ld -R -o"

#define PATHSEP '\\'
#define EXT	"ext_s enter_s"
#define CRT	"crt"
#define SWEXT	"swext_s enter_s"
#define SWCRT	"swcrt"
#define SCEXT	"swext_s start_s"
#define COMP	"compile.bat"
#define EXEC	"compile"
#else
#include "local.h"
#define PATHSEP '.'
#endif

#define LIT	"litfile"
#define OFILE	"a/out"

#define TRUE 1
#define FALSE 0
#define SAME 0				/* strcmp() equal */
#define UNDEFINED NULL			/* file status */
#define RW 3				/* MOS read/write mode */

char tmp1[128];			/* compiler temporaries */
char tmp2[128];
char tmp3[128];
char tmp4[128];
char pathname[128];			/* pathname of source file */
char basename[128];			/* pathname with suffix removed */
char suffix;				/* file type 'c' suffix */
char *pp;				/* pathname pointer */
char *bp;				/* basename pointer */
char *exec;				/* invocation mantra */
char *tcpp;				/* Small-C preprocessor */
char *tccom;				/* Small-C compiler */
char *as65;				/* 6502 assember */
char *ld;				/* 6502 loader */
char *rm;				/* file remover */
char *ext;				/* external references */
char *crt;				/* Small-C run-time support module */
char *lit;				/* compiler literal pool */
char *ofile;				/* output object file */
char *comp;				/* exec file */
FILE *in, *out;			/* input/output files */
int doexec;				/* execute *exec file */
int docom;				/* compiler phase */
int doasm;				/* assembly phase */
int doload;				/* load phase */
int keepsym;				/* preserve global symbol file */

int exhdl;				/* JGH: exec file handle */

int main(argc, argv)
int argc;
char *argv[];
{
	in = UNDEFINED;
	out = UNDEFINED;
	if (argc < 2)
		usage();
	doexec = TRUE;
	docom = TRUE;
	doasm = TRUE;
	doload = TRUE;
	keepsym = FALSE;
	tcpp = TCPP;
	tccom = TCCOM;
	as65 = AS65;
	rm = RM;
	ld = NULL;
	crt = NULL;
	ext = SCEXT;
	lit = LIT;
	ofile = OFILE;
	comp = COMP;
	exec = EXEC;
	exhdl=0;
	while ((*argv[1] & 0xFF) == '-') {	/* BUG in compiler */
		if (strcmp(argv[1], "-o") == SAME) {
			if (argc < 3)
				usage();
			else {
				ofile = argv[2];
				--argc;
				argv++;
			}
		}
		else if (strcmp(argv[1], "-n") == SAME)
			--doexec;
		else if (strcmp(argv[1], "-E") == SAME) {
			--docom;
			--doasm;
		}
		else if (strcmp(argv[1], "-S") == SAME)
			--doasm;
		else if (strcmp(argv[1], "-c") == SAME)
			--doload;
		else if (strcmp(argv[1], "-R") == SAME) {
			ld = SWLD;
			ext = SWEXT;
			crt = SWCRT;
		}
		else if (strcmp(argv[1], "-A") == SAME) {
			ld = LD;
			ext = EXT;
			crt = CRT;
		}
		else if (strcmp(argv[1], "-C") == SAME)
			tccom = TCCOMC;
		else if (strcmp(argv[1], "-g") == SAME)
			keepsym++;
		else
			usage();
		--argc;
		argv++;
	}
	if (argc < 2)
		usage();
	strcpy(pathname, argv[1]);
	pp = pathname;
	while (*pp++)
		;
	while (--pp != pathname) {
		if (*pp == PATHSEP) {
			++pp;
			break;
		}
	}
	bp = basename;
	while ((*bp = *pp++) != '\0')
		++bp;
	suffix = *--bp;
	*--bp = '\0';
	if (suffix != 'c') {
		fprintf(stderr, "tcc: can't compile '%s'\n", pathname);
		fatal(-1);
	}
	if ((in = fopen(pathname, "r")) == NULL) {
		fprintf(stderr, "tcc: can't open %s\n", pathname);
		fatal(-1);
	}
	fclose(in);
	in = UNDEFINED;

	exhdl=(osbyte(198, 0) & 0xFF);
		/* Find current exec handle and disconnect input */
	if ((out = fopen(comp, "w")) == NULL) {
		fprintf(stderr, "tcc: can't open %s\n", comp);
		fatal(-1);
	}
	sprintf(tmp1, "%s/i", basename);
	sprintf(tmp2, "%s/s", basename);
	sprintf(tmp3, "%s/o", basename);
	sprintf(tmp4, "g/out");
	if (ld == NULL) {
		strcpy(tmp3, ofile);
		strcpy(ofile, "");
	}
	fprintf(out, "%s %s %s %s %s ", rm, tmp1, tmp2, tmp3, tmp4);
	fprintf(out, "%s %s\n", lit, ofile);
	fprintf(out, "%s %s %s\n", tcpp, pathname, tmp1);
	if (docom) {
		fprintf(out, "%s %s %s\n", tccom, tmp1, tmp2);
		fprintf(out, "%s %s %s\n", rm, tmp1, lit);
	}
	if (doasm) {
		fprintf(out, "%s %s %s %s\n", as65, tmp3, ext, tmp2);
		if (keepsym)
			fprintf(out, "%s %s\n", rm, tmp2);
		else
			fprintf(out, "%s %s %s\n", rm, tmp2, tmp4);
		if ((ld != NULL) & doload) {
			fprintf(out, "%s %s %s %s\n", ld, ofile, crt, tmp3);
			fprintf(out, "%s %s\n", rm, tmp3);
		}
	}
	if (exhdl)
		fprintf(out, "*fx 198,%d\n", exhdl);
		/* Reconnect to exec */
	else
		fprintf(out, "%s %s\n", rm, comp);
		/* Remove compile batchfile */
	fclose(out);
	if (doexec)
		system(exec);
	exit(0);
}

/*
 * report usage and exit
 */
usage()
{
	fprintf(stderr, "usage: tcc [-o outfile] [-n] [-E] [-S] [-c] [-g]");
	fprintf(stderr, " [-R] [-A] [-C] file\n");
	fatal(-1);
}

/*
 * close files and exit
 */
fatal(stat)
int stat;				/* exit status */
{
	if (in != NULL)
		fclose(in);
	if (out != NULL)
		fclose(out);
	exit(stat);
}
